/****************************************************************************
*****************************************************************************
**                                                                         **
**   The Ghostscript extender by Alan Kreutzer, August 2003.               **
**                                                                         **
**      Provides an interface between WinBatch and the Ghostscript API.    **
**                                                                         **
**      This program has been tested against AFPL Ghostscript 8.0          **
**                                                                         **
*****************************************************************************
****************************************************************************/
#define INTERFACELEVEL  2           // extender interface level 
#define EXTENDER_FEATURE_LEVEL 1001 // extender feature level 

#define STRICT

/*-----------------------------------------------------------------------+
!   When SHOWALLOC = 1 we display all Allocate and Free events.          !
!   When SHOWALLOCTOTAL = 1 we count all Allocate and Free events,       !
!                           and display a total at shutdown.             !
!   This is for debug to detect memory leaks.                            !
+-----------------------------------------------------------------------*/ 
#define SHOWALLOC 0
#define SHOWALLOCTOTAL 0


/*****************************************************************************
*                                                                            *
*                                   Includes                                 *
*                                                                            *
*****************************************************************************/
#include <windows.h>
#include <stdio.h>
                           /* special includes for WinBatch extenders    */
#include "wwwpform.h"
#include "addons.h"


/*******************************************************************************
*                                                                              *
*   This structure is used store information between calls.                    *
*                                                                              *
*   The "GLOBAL" structure is created when the extender is loaded.             *
*   A pointer to the "GLOBAL" structure is saved in the "lpLong" parameter.    *
*                                                                              *
*******************************************************************************/
typedef struct _GLOBAL
{ 
     char        ReturnType;    /* Return type S=String,L=Long,E=Error           */
     long        ErrorType;     /* Error type... -1 (Normal) or -3 (Fatal)       */	  
     long        Long;          /* Numeric value (returned when Type='L')        */
     char       *String;        /* String        (Returned when Type='S')        */    
     char        Buffer[256];   /* Buffer for messages. (Returned when Type='E') */
 } GLOBAL;

/*****************************************************************************
*                                                                            *
*                Definitions for the Ghostscript API                         *
*                                                                            *
*****************************************************************************/
#define GSDLLAPI     __stdcall
#define GSDLLCALL    __stdcall
#define GSDLLCALLPTR GSDLLCALL * 

typedef  long  gs_main_instance;

typedef struct gsapi_revision_s {
    const char *product;
    const char *copyright;
    long       revision;
    long       revisiondate;
} gsapi_revision_t;


/*****************************************************************************
*                                                                            *
*                           Function Prototypes                              *
*                                                                            *
*****************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved);

HGLOBAL Allocate (GLOBAL *Global, long Size, char *Name); 
void    Free     (void *Memory, char *Name);  
void    MBSS     (char *Title, char *Text);  
void    MBSN     (char *Title, long Number);
void    MBSX     (char *Title, long Number);

long ExtenderVersion    ();
void LoadMessage        (long MsgID, char *Buffer, char *Param1, char *Param2);
char *FindFileName      (char *Path);
BOOL LoadGhostscriptDLL (GLOBAL *Global);
void *FindCall          (GLOBAL *Global, HINSTANCE hDLL, char *Library, char *EntryPoint);
HINSTANCE FindLibrary   (GLOBAL *Global, char *Library);

#define LOWERCASE(p1) LOBYTE(LOWORD((DWORD) (CharLower((LPSTR) (DWORD) (BYTE) (p1)))))

/*****************************************************************************
*                                                                            *
*                           Global Variables                                 *
*                                                                            *
*****************************************************************************/
         /*  Superstitious data field to bypass/ignore bugs in certain versions
         !   of Windows that cause the beginning of the data area to be
         !  overwritten.  Should be the first data area defined and linked. */
char        szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE   hinstDLL;   // DLL hInstance saved here    

/*-----------------------------------------------------------------------+
!                         For debugging                                  !
+-----------------------------------------------------------------------*/
#if SHOWALLOCTOTAL == 1
long DebugAllocateCounter = 0;
#endif

/*****************************************************************************
*                                                                            *
*              Pointers to functions in the ghostscript dll.                 *
*                                                                            *
*****************************************************************************/
HINSTANCE   hGSdll;
int  GSDLLAPI (*gsapi_revision)       (gsapi_revision_t *pr, int len);
int  GSDLLAPI (*gsapi_new_instance)   (gs_main_instance **pinstance, void *caller_handle);
void GSDLLAPI (*gsapi_delete_instance)(gs_main_instance *instance);
int  GSDLLAPI (*gsapi_init_with_args) (gs_main_instance *instance, int argc, char **argv);
int  GSDLLAPI (*gsapi_exit)           (gs_main_instance *instance);
int  GSDLLAPI (*gsapi_set_stdio)      (gs_main_instance *instance,
    int (GSDLLCALLPTR stdin_fn) (void *caller_handle, char *buf, int len),
    int (GSDLLCALLPTR stdout_fn)(void *caller_handle, const char *str, int len),
    int (GSDLLCALLPTR stderr_fn)(void *caller_handle, const char *str, int len));
    
static int GSDLLCALL gsdll_stdin(void *instance, char *buf, int len);
static int GSDLLCALL gsdll_stdout(void *instance, const char *str, int len);
static int GSDLLCALL gsdll_stderr(void *instance, const char *str, int len);


/*****************************************************************************
*                                                                            *
*            This global area is used by the callback routines               *
*            to save the messages that GhostScript writes to                 *
*            "stdout" and "errout"                                           *
*                                                                            *
*****************************************************************************/
long global_stdoutIndex;       // Index to text in the "stdout" buffer;
char global_stdout[2048];      // Buffer receives text sent to "stdout"
     

/*****************************************************************************
*                                                                            *
*      Command Extension Table for this DLL                                  *
*                                                                            *
*****************************************************************************/
tablestruct CommandExtTable[] =
{
/*-----------------------------------------------------------------------------+
!  ID   Type       Mask            Params/Value, reset,handle,Name             !
+------------------------------------------------------------------------------*/

  {10, FUNCTION, 0x0000000000000000, 0x00000001L, 0, "ghostscriptExtenderInfo" },
  {11, FUNCTION, 0x0000000000000000, 0x00000000L, 0, "ghostscriptVersion"      },  
  {12, FUNCTION, 0x0000000000000000, 0x00000000L, 0, "ghostscriptMessages"     },
  {13, FUNCTION, 0x1111111111111111, 0x00000F01L, 0, "ghostscriptCommand"      },
  {14, FUNCTION, 0x0000000000000011, 0x00000002L, 0, "ghostscriptPS2PDF"       },
  {15, FUNCTION, 0x0000000000000011, 0x00000201L, 0, "ghostscriptPrint"        },  
/*-----------------------------------------------------------------------------+
!                           Table ends with ID = -1                            !
+------------------------------------------------------------------------------*/
  {-1,        0, 0x0000000000000000, 0x00000000L,  0, ""                  }
};


/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                               DllMain                                      **
**                                                                            **
**                                                                            **
**   Called when the DLL is loaded or unloaded.                               **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{   
   if (dwReason == DLL_PROCESS_ATTACH)
   {
      hinstDLL = hInst;
      global_stdoutIndex = 0;
      global_stdout[0] = '\0';    
   }

#if SHOWALLOCTOTAL == 1
   if (dwReason == DLL_PROCESS_DETACH)
   {
       if (DebugAllocateCounter)
       {
          char Work[256];
          wsprintf(Work,"<<< WARNING >>> There are %d Allocated Blocks Remaining!  <<< WARNING >>>",DebugAllocateCounter);
          MBSS("XFTP34I.DLL is Unloading",Work);
       }
       else
       {
          MBSS("XFTP34I.DLL is Unloading","All allocated blocks have been freed.");
       }
   }
#endif
 
   return (TRUE);
   UNREFERENCED_PARAMETER(lpReserved);
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                        WILExtenderQuery                                    **
**                                                                            **
**           Tells the Wil interpereter about our interface level.            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{

  char  szWork[128];
  LONG  i;

  switch (nRequest)
  {
  
/*----------------------------------------------------------------------+
!   case 1 ... identify yourself to a caller.                           !
+----------------------------------------------------------------------*/
    case 1:                      
    {
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }
      return ExtenderVersion();
    }

/*----------------------------------------------------------------------+
!   case 2 ... Tell WinBatch we use interface level 2                   !   
+----------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }

/*----------------------------------------------------------------------+
!    unknown requests                                                   !  
+----------------------------------------------------------------------*/
    default:
      return (0);
  }
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                          WILExtender2                                      **
**                                                                            **
**                                                                            **
**                                                                            **
**             The main routine performs all extender functions.              **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/

DllExport LONG FAR PASCAL WILExtender2
  (HWND           hWnd,                   // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{     
  long            Ret;
  GLOBAL          Global;

  Global.ReturnType = ' ';
  
     /***************************************************
     *                                                  *
     *    The "Big Switch"                              *
     *                                                  *
     *      Initialization calls.                       *
     *                                                  *
     ***************************************************/             
  switch (CmdId)         
  {
  
    UNREFERENCED_PARAMETER(hWnd);
    UNREFERENCED_PARAMETER(lpLong);
    
/*======================================================================+
!  case -1 ... One time initialization routine.                         !
!                                                                       !
!              Return the number of functions  & constants defined,     !
!              or return -1 if the initialization fails.   lpvvArg      !
!              points to a character buffer to receive an error message !
!              and nArgCount contains the size of the buffer.           !
+======================================================================*/
    case -1:                         
    {  
       long X;
       

       LoadGhostscriptDLL(&Global);
       if (Global.ReturnType == 'E')
       {
            lstrcpyn((char*)lpvvArg,Global.Buffer,nArgCount);
            return -1;
       }
       
       /*----------------------------------------------------------+
       !  Normal return.  Number of constants and functions.       !
       +----------------------------------------------------------*/
       X = 0;
       while (1)
       {
           if (CommandExtTable[X].ident == -1) break;
           X++;
       }   
       return X; 
    }

/*======================================================================+
!  case -2 ... Called repeatedly as extender is being loaded.           !
!                                                                       !
!              On each call, return a function table entry.             !
+======================================================================*/
    case -2:                          
    {
      LPTABLESTRUCT lpTable;

      if (CommandExtTable[nArgCount].ident == -1)
        return (0);
      
      /*-----------------------------------------------------------------+
      !   Pass entry point infomation back to the WIL interpereter.      !
      +-----------------------------------------------------------------*/
      lpTable = (LPTABLESTRUCT) lpvvArg;
      if (lpTable)
      {
        lpTable->ident  = CommandExtTable[nArgCount].ident;
        lpTable->type   = CommandExtTable[nArgCount].type;
        lpTable->param1 = CommandExtTable[nArgCount].param1;
        lpTable->ulMask = CommandExtTable[nArgCount].ulMask;
        lstrcpy((LPSTR) lpTable->delim, (LPSTR) CommandExtTable[nArgCount].delim);       
        CharLower(lpTable->delim);
        lpTable->size   = lstrlen(lpTable->delim);
      }

      return (1);
    }


/*======================================================================+
!  case -3 .. Called once at shutdown. Can be used for cleanup.         !
+======================================================================*/
    case -3:
    {
        FreeLibrary(hGSdll); // Free the Ghostscript DLL
        return (1);
    }

/*======================================================================+
!  case 10  ... ghostscriptExtenderInfo                                 !
!                                                                       !
!              Return extender information.                             !
!                                                                       !
!              Parameter 0 = Version                                    !
!                        1 = Total Functions                            !
!                        2 = Total Constants                            !
+======================================================================*/
    case 10:
    {       
         UINT uType;
         long X;
         
         if (lpvvArg[0].x == 0)           /* Check the request number */
         {
              Global.ReturnType = 'L';
              Global.Long = ExtenderVersion();
              break;
         } 
   
         if (lpvvArg[0].x == 1)            /* total functions         */
             uType = FUNCTION;
         else 
            if (lpvvArg[0].x == 2)         /* total constants         */
                uType = CONSTANT;
            else
            {
                Global.ReturnType = 'E';   /* Invalid parameter      */
                Global.ErrorType = -1;
                LoadMessage(2007,Global.Buffer,"","");                
                break;
            }

        Global.ReturnType = 'L';
        Global.Long = 0;
        for (X = 0; CommandExtTable[X].ident != -1; X++)
           if ((UINT) CommandExtTable[X].type == uType)
                Global.Long++;
        break;
    }

/*======================================================================+
!  case 11 ... ghostscriptVersion()                                     !
!                                                                       !
!   Returns the version Ghostscript we're using.                        !
+======================================================================*/ 
    case 11:
    {        
       gsapi_revision_t GSRevInfo;
    
       (*gsapi_revision)(&GSRevInfo, sizeof(GSRevInfo));
       Global.ReturnType = 'L';
       Global.Long = GSRevInfo.revision;
       break;
    }
    
/*======================================================================+
!  case 12 ... ghostscriptMessages ()                                   !
!                                                                       !
!   Returns the text that ghostscript wrote to "stdout" or "stderr"     !
!   while processing the last command.  For successful commands there   !
!   is usualy no need to see this message,  but if there was an error   !
!   the messages may explain why.                                       !
+======================================================================*/ 
    case 12:
    {
        Global.ReturnType = 'S';
        Global.String = global_stdout;
        break;
    }

/*=========================================================================+
!   case 13 ... ghostscriptCommand(Parameters)                             !
!                                                                          !
!    The parameters are any parameters you could use on a ghostscript      !
!    command line.   The resulting command is executed.                    !
+=========================================================================*/ 
    case 13:
    {
          gs_main_instance *InstanceHandle;
          char   *gsargv[20];
          int     gsargc;
          long    X;
 
          //----------------------------------------------------------
          // Open a Ghostscript instance
          //----------------------------------------------------------         
          Ret = (*gsapi_new_instance)(&InstanceHandle, NULL);
          if (Ret < 0)
          {
             char N[24];
             Global.ReturnType = 'E';
             Global.ErrorType = -1;
             wsprintf(N,"%d",Ret);
             LoadMessage(2008,Global.Buffer,N,"");
             break;
          }

          //----------------------------------------------------------
          // Build a parameter list.
          //---------------------------------------------------------- 
          gsargv[0] = "gs";        /* actual value doesn't matter      */
          gsargv[1] = "-dNOPAUSE"; /* suppress prompt for next command */
          gsargv[2] = "-q";        /* only return error messages.      */
          for (X=0,gsargc=3;X<nArgCount;X++,gsargc++)
          {
             gsargv[gsargc] = lpvvArg[X].lpstr;
          }      
          
         //----------------------------------------------------------
          // Call the Ghostscript routines to do the conversion
          //----------------------------------------------------------
          lstrcpy(global_stdout,"Done.");
          global_stdoutIndex = 0;

          (*gsapi_set_stdio)(InstanceHandle, gsdll_stdin, gsdll_stdout, gsdll_stderr);
          Ret = (*gsapi_init_with_args)(InstanceHandle, gsargc, gsargv);
          (*gsapi_exit)(InstanceHandle);
          (*gsapi_delete_instance)(InstanceHandle);
          Global.ReturnType = 'L';
          if (Ret < 0)
              Global.Long = Ret;
          else 
              Global.Long = 0;       
          break;
    }  

    
    
/*=========================================================================+
!   case 14 ... ghostscriptPS2PDF(InputFile, OutputFile )                  !
!                                                                          !
!    Convert a Postscript file to PDF.                                     !
!    Returns 1 if successful, or 0 if there was an error.
+=========================================================================*/ 
    case 14:
    {
          gs_main_instance *InstanceHandle;
          char   *gsargv[12];
          int     gsargc;
          long    X;
          char   *OutputParameter;
 
          //----------------------------------------------------------
          // Open a Ghostscript instance
          //----------------------------------------------------------         
          Ret = (*gsapi_new_instance)(&InstanceHandle, NULL);
          if (Ret < 0)
          {
             char N[24];
             Global.ReturnType = 'E';
             Global.ErrorType = -1;
             wsprintf(N,"%d",Ret);
             LoadMessage(2008,Global.Buffer,N,"");
             break;
          }

          //----------------------------------------------------------
          // We are passed the output file name as a parameter.
          // We need to pass "-sOutputFile=OutputFileName".
          // Borrow some memory and build this value.
          //----------------------------------------------------------
          X = lstrlen(lpvvArg[1].lpstr) + 20;
          OutputParameter = Allocate(&Global,X,"OutFile Parameter");
          if (Global.ReturnType == 'E')  break;
          lstrcpy(OutputParameter,"-sOutputFile="); 
          lstrcat(OutputParameter,lpvvArg[1].lpstr);

          //----------------------------------------------------------
          // Build a parameter list.
          //----------------------------------------------------------         
          gsargv[0]  = "ps2pdf";      /* actual value doesn't matter */
          gsargv[1]  = "-dNOPAUSE";
          gsargv[2]  = "-dBATCH";
          gsargv[3]  = "-dSAFER";
          gsargv[4]  = "-q";
          gsargv[5]  = "-dSHORTERRORS";
          gsargv[6]  = "-sDEVICE=pdfwrite";
          gsargv[7]  = OutputParameter;
          gsargv[8]  = "-c";
          gsargv[9]  = ".setpdfwrite";
          gsargv[10]  = "-f";
          gsargv[11] = lpvvArg[0].lpstr;
          gsargc=12;

         //----------------------------------------------------------
          // Call the Ghostscript routines to do the conversion
          //----------------------------------------------------------
          lstrcpy(global_stdout,"Done.");
          global_stdoutIndex = 0;

          (*gsapi_set_stdio)(InstanceHandle, gsdll_stdin, gsdll_stdout, gsdll_stderr);
          Ret = (*gsapi_init_with_args)(InstanceHandle, gsargc, gsargv);
          (*gsapi_exit)(InstanceHandle);
          (*gsapi_delete_instance)(InstanceHandle);
          Free(OutputParameter,"OutFile Parameter");
          Global.ReturnType = 'L';
          if (Ret < 0)
              Global.Long = Ret;
          else 
              Global.Long = 0;       
          break;
    }  

    
/*==========================================================================+
!   case 15 ... ghostscriptPrint(InputFile, Printer, StatusWindow)          !
!                                                                           !
!    Send a postscript or PDF file to a windows printer.                    !
!    if "StatusWindow" = TRUE we show a status window with a CANCEL button. !
+==========================================================================*/ 
    case 15:
    {
          gs_main_instance *InstanceHandle;
          char   *Printer;
          char   *PrinterParameter;
          long    Window;
          char   *gsargv[12];
          int     gsargc;
          long    X;

          //----------------------------------------------------------
          // Default values if optional arguments are omitted.
          //----------------------------------------------------------           
          if (nArgCount > 1)  Printer = lpvvArg[1].lpstr;
          else                Printer = "";
              
          if (nArgCount > 2)  Window = lpvvArg[2].x;
          else                Window = TRUE;

          //----------------------------------------------------------
          // Open a Ghostscript instance
          //----------------------------------------------------------         
          Ret = (*gsapi_new_instance)(&InstanceHandle, NULL);
          if (Ret < 0)
          {
             char N[24];
             Global.ReturnType = 'E';
             Global.ErrorType = -1;
             wsprintf(N,"%d",Ret);
             LoadMessage(2008,Global.Buffer,N,"");
             break;
          }

          //----------------------------------------------------------
          // We are passed the printer  name as a parameter.
          // We need to pass "-sOutputFile="%printer%PrinterName".
          // Borrow some memory and build this value.
          //----------------------------------------------------------
          X = lstrlen(Printer) + 32;
          PrinterParameter = Allocate(&Global,X,"Printer Parameter");
          if (Global.ReturnType == 'E')  break;
          lstrcpy(PrinterParameter,"-sOutputFile=%printer%"); 
          lstrcat(PrinterParameter,Printer);

          //----------------------------------------------------------
          // Build a parameter list.
          //----------------------------------------------------------         
          gsargv[0] = "gs";	       /* actual value doesn't matter */
          gsargv[1] = "-dNOPAUSE";
          gsargv[2] = "-dBATCH";
          gsargv[3] = "-dSAFER";
          gsargv[4] = "-dSHORTERRORS";
          gsargv[5] = "-q";
          gsargv[6] = "-sDEVICE=mswinpr2";
          gsargv[7] = PrinterParameter;
          
          if (Window == 0)
          {
             gsargv[8] = "-dNoCancel";
             gsargc = 9;
          }
          else
          {
             gsargc = 8;
          }
          gsargv[gsargc] = lpvvArg[0].lpstr;
          gsargc++;

          //----------------------------------------------------------
          // Call the Ghostscript routines to do the conversion
          //----------------------------------------------------------
          lstrcpy(global_stdout,"Done.");
          global_stdoutIndex = 0;
           
          (*gsapi_set_stdio)(InstanceHandle, gsdll_stdin, gsdll_stdout, gsdll_stderr);
          Ret = (*gsapi_init_with_args)(InstanceHandle, gsargc, gsargv);
          (*gsapi_exit)(InstanceHandle);
          (*gsapi_delete_instance)(InstanceHandle);
          Free(PrinterParameter,"Printer Parameter");
          Global.ReturnType = 'L';
          if (Ret < 0)
              Global.Long = Ret;
          else 
              Global.Long = 0;       
          break;
    }  

/*======================================================================+
!   Default Case.  (We should never get here)                           !
+======================================================================*/
    default:      
    {
        Global.ReturnType = 'E';
        Global.ErrorType = -3;
        LoadMessage(2001,Global.Buffer,"","");  
        break;
    }
  }                   
  
  
            /*************************************
            *                                    *
            *       End of "Big Switch"          *
            *                                    *
            *************************************/
 
/*-----------------------------------------------------------------------+
!+----------------------------------------------------------------------+!
!!                                                                      !!
!!   Return the value to the caller.                                    !!
!!                                                                      !!
!+----------------------------------------------------------------------+!
+-----------------------------------------------------------------------*/    

  // --------------------------------------------------------------
  //   Return a value to the user
  // --------------------------------------------------------------
  switch (Global.ReturnType)       
  { 
  
    case 'L':  // Return a long integer
    {
       lpViperReturn->inittype=VARTYPE_INT;
       lpViperReturn->x=Global.Long;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
    
    case 'S':   // Return a string
    {
       lpViperReturn->inittype=VARTYPE_STRING;
       lpViperReturn->lpstr=Global.String;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);    
       break;
    }

    case 'E':  // Return an error message
    {
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = Global.Buffer;
       Ret = lpViperCallback(Global.ErrorType, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
    
    default:  // Invalid return type.  (This better not happen!)
    {
       char Work[128];
       LoadMessage(2002, Work, "", "");
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = Work;
       Ret = lpViperCallback(-3, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }  
  }

  return Ret;
  
  UNREFERENCED_PARAMETER(hInst);
}
                 /*************************************
                 *                                    *
                 *    End of program WILExtender2     *
                 *                                    *
                 *************************************/




/***************************************************************************
*                                                                          *
*                                                                          *
*              Allocate and Free Memory                                    *
*                                                                          *
*                                                                          *
***************************************************************************/

     /*------------------------------------------------------------------+
     !   Allocate memory.                                                !
     +------------------------------------------------------------------*/
HGLOBAL Allocate(GLOBAL *Global, long Size, char *Name)
{
     HGLOBAL Mem;
     Mem = GlobalAlloc(GPTR,Size);
     if (Mem != NULL)
     {
#if SHOWALLOC == 1
        char Work[256];
        wsprintf(Work,"Allocate x%08lX -> %s",Mem,Name); 
        MBSS("Allocate",Work);
#endif
#if SHOWALLOCTOTAL == 1
        DebugAllocateCounter++;
#endif     
        return Mem;
     }
     Global->ReturnType = 'E';
     Global->ErrorType = -3;
     LoadMessage(2003,Global->Buffer,"","");
     return NULL;
     UNREFERENCED_PARAMETER(Name);
}

     /*------------------------------------------------------------------+
     !   Free memory.                                                    !
     +------------------------------------------------------------------*/
void Free(void* Memory, char *Name)
{
    HGLOBAL Result;
    
    if (Memory == NULL) return;   // NULL = We didn't borrow any memory.
    Result = GlobalFree(Memory);  // Free the memory.

#if SHOWALLOC == 1
    if (Result == (HGLOBAL) NULL)
    {
       char Work[256];
       wsprintf(Work,"Free x%08lX  -> %s",Memory,Name);
       MBSS("Free",Work);
    }
    else
    {
       char Work[256];
       wsprintf(Work,"Free FAILED x%08lX  -> %s",Memory,Name);
       MBSS("FAILED",Work);    
    }
#endif  

#if SHOWALLOCTOTAL == 1
    if (Result == (HGLOBAL) NULL)
    {
       DebugAllocateCounter--;
    }
#endif

    return; 
    UNREFERENCED_PARAMETER(Name);
} 


 
 
/***************************************************************************
*                                                                          *
*     Debug Functions                                                      *
*                                                                          *
*      MBSN ... Display a decimal number.                                  *
*      MBSX ... Display a hexidecimal number.                              *
*      MBSS ... Display a string.                                          *
*      MBLL ... Display a long long (64 bit) integer.                      *
*                                                                          *
***************************************************************************/ 
 void MBSN(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"%d",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }

 void MBSX(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"x%08lX",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void MBSS(char *Title, char *Text)
{
     MessageBox(GetFocus(), Text, Title, MB_OK | MB_SETFOREGROUND);
}



/*****************************************************************************
*                                                                            *
*   Get the extender version from the resource file, and return it as        *
*   an integer.                                                              *
*                                                                            *
*****************************************************************************/
long ExtenderVersion()
{
    char Work[24];
    char *Wx;
    char C;
    long Version;
    
    LoadString(hinstDLL, 1001, Work, sizeof(Work));
    Wx = Work;
    
    Version = 0;
    while ((C = *Wx++) != '\0') Version = (Version * 10) + (C - '0');   
    return Version;
}

/****************************************************************************
*  Load a message from the resource file.                                   *
*                                                                           *
*    There are two insertable parameters.  (which replace %s where          *
*    it appears in the message.)   The message is returned in Buffer,       *
*    which is assumed to be 256 bytes long ... long enough for any message  *
*    in our resource file.                                                  *
*                                                                           *
*    Originally we used WSPRINTF to do the %s substitution.  However        *
*    WSPRINTF does not check for buffer overflow, so it was replaced        *
*    with the routine below.                                                *
*                                                                           *
****************************************************************************/
void LoadMessage(long MsgID, char *Buffer, char *Param1, char *Param2)
{
    char Work[256];
    char *Wx;
    char *Bx;
    char *Px;
    char *P2x;
    long L;
    
    LoadString(hinstDLL, MsgID, Work, sizeof(Work));
    
    Wx = Work;
    Bx = Buffer;
    Px = Param1;
    P2x = Param2;
    L = 0;
    while (1)
    {
        if (*Wx == '%')
        {
          if (LOWERCASE(*(Wx+1)) == 's')
          {
              Wx +=2;
              while (1)
              {
                  if (*Px == '\0') break;
                  L++;
                  if (L == 256) {*Bx = '\0'; return;}             
                  *Bx++ = *Px++;
              }
              Px = P2x;
              P2x = "";
              continue;
          } 
        }  
        L++;
        if (L == 256) {*Bx = '\0'; return;}  
        *Bx++ = *Wx++;
    }
    return;
}     



/*************************************************************************************
*                                                                                    *
*                                                                                    *
*    Find ghostscript DLL and get the addresses of the entry points we need.         *
*                                                                                    *
*                                                                                    *
*************************************************************************************/
BOOL  LoadGhostscriptDLL(GLOBAL *Global)
{
    char Library[] = "gsdll32.dll";
    
    hGSdll = FindLibrary(Global, Library); 
    if (Global->ReturnType == 'E') return FALSE;

    (void *) gsapi_revision        = FindCall(Global,hGSdll,Library,"gsapi_revision");       
    (void *) gsapi_new_instance    = FindCall(Global,hGSdll,Library,"gsapi_new_instance"); 
    (void *) gsapi_delete_instance = FindCall(Global,hGSdll,Library,"gsapi_delete_instance");       
    (void *) gsapi_init_with_args  = FindCall(Global,hGSdll,Library,"gsapi_init_with_args"); 
    (void *) gsapi_exit            = FindCall(Global,hGSdll,Library,"gsapi_exit"); 
    (void *) gsapi_set_stdio       = FindCall(Global,hGSdll,Library,"gsapi_set_stdio"); 
    

    if (Global->ReturnType == 'E')
    {
       FreeLibrary(hGSdll);  
       hGSdll = NULL;
       return FALSE;
    }

    return (TRUE);
}
 
 
/***************************************************************************
*                                                                          *
*   Given a file path, find the file name at the end of the path.          *
*                                                                          *
***************************************************************************/
char *FindFileName(char *Path)
{
    char *FileName;
    char C;

    FileName = Path;
    while (1)
    {
        C = *Path++;
        if (C == '\0') return FileName;
        if (C == '\\' || C == '/') FileName = Path;
    }
}
/*****************************************************************************
*                                                                            *
*    Load a DLL library.                                                     *
*                                                                            *
*****************************************************************************/  
HINSTANCE FindLibrary(GLOBAL *Global, char *Library)
{
    HINSTANCE  Handle;
    char       Path[MAX_PATH+4];
    char       *F;
    long       X;
    long       L;
        
    X = GetModuleFileName(hinstDLL,Path,MAX_PATH+1);  // Get this DLL's path.
    if (X > 0)
    {
        F = FindFileName(Path);                       // Look for the new DLL 
        L = MAX_PATH - (F - Path);
        lstrcpyn(F,Library,L);                        // in the same path.
        Handle = LoadLibrary(Path);
        if (Handle != NULL) return Handle;
    }

    Handle = LoadLibrary(Library);                    // Next, look in user's directories
    if (Handle == NULL) 
    { 
       Global->ReturnType = 'E';
       Global->ErrorType = -3;
       LoadMessage(2004, Global->Buffer, Library, ""); 
       return NULL;
    }        
    return Handle;
}
    
/*****************************************************************************
*                                                                            *
*    Find the address of a particular entry point in a DLL.                  *
*                                                                            *
*****************************************************************************/
void *FindCall(GLOBAL *Global, HINSTANCE hDLL, char *Library, char *EntryPoint)
{
     void *Address;
     Address = GetProcAddress(hDLL,EntryPoint);
     if (Address == NULL)
     {
       Global->ReturnType = 'E';
       Global->ErrorType = -3;
       LoadMessage(2006, Global->Buffer, EntryPoint, Library); 
    }             
    return Address;
}



/*****************************************************************************
*                                                                            *
*    Callback Function for STDIN.                                            *
*                                                                            *
*****************************************************************************/
static int GSDLLCALL gsdll_stdin(void *instance, char *buf, int len)
{
    return 0;      // We don't have a console, so just return EOF
    
    UNREFERENCED_PARAMETER(instance);
    UNREFERENCED_PARAMETER(buf);    
    UNREFERENCED_PARAMETER(len);
}

/*****************************************************************************
*                                                                            *
*    Callback Function for STDOUT.                                           *
*     Saves all text in "global_stdout"                                      *
*                                                                            *
*****************************************************************************/
static int GSDLLCALL gsdll_stdout(void *instance, const char *str, int len)
{
   long ToCopy;
   long Max;
   
   Max = sizeof(global_stdout) - global_stdoutIndex;
   ToCopy = len+1;
   if (ToCopy > Max) ToCopy = Max;
   lstrcpyn(&global_stdout[global_stdoutIndex],str,ToCopy);
   fflush(stdout);
   global_stdoutIndex += len;
   return len;
   
   UNREFERENCED_PARAMETER(instance);
}

/*****************************************************************************
*                                                                            *
*    Callback Function for ERROUT.   (Passes info to the STDOUT callback.)   *
*                                                                            *
*****************************************************************************/
static int GSDLLCALL gsdll_stderr(void *instance, const char *str, int len)
{
    return gsdll_stdout(instance, str, len);
}
